import logging

class AnilistFormatter(logging.Formatter):

    black = "\x1b[30;20m"
    blue = "\x1b[34;20m"
    green = "\x1b[32;20m"
    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "[%(asctime)s]: %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"

    FORMATS = {
        logging.DEBUG: green + format + reset,
        logging.INFO: blue + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)
    
# Singleton
class AnilistLogger:
    _instance = None
    _done = False

    def __init__(self, level = logging.INFO):
        if not AnilistLogger._done:
            logger = logging.getLogger("AnilistLogger")
            logger.setLevel(level)

            ch = logging.StreamHandler()
            ch.setLevel(level)

            ch.setFormatter(AnilistFormatter())

            logger.addHandler(ch)

            self._logger = logger
            AnilistLogger._done = True

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(AnilistLogger, cls).__new__(cls)
        
        return cls._instance

    def __getattr__(self, k):
        return self._logger.__getattribute__(k)


class AnilistObject:
    
    def __init__(self, json_object):
        self.__dict = {}
        for k, v in json_object.items():
            if type(v) is dict:
                self.__dict[k] = AnilistObject(v)
            elif type(v) is list:
                self.__dict[k] = [AnilistObject(el) if type(el) is dict else el for el in v]
            else:
                self.__dict[k] = v

    def __getattr__(self, k):
        return self.__dict[k]
    
    def __str__(self):
        return str(self.__dict)
    
    def __repr__(self):
        return str(self)


import socket
import re
import webbrowser

pat = re.compile(r'\?code=(.*?)(\&|\r| |\n)')

def get_code(client_id):
    webbrowser.open(f"https://anilist.co/api/v2/oauth/authorize?client_id={client_id}&redirect_uri=http://127.0.0.1:8000/&response_type=code", autoraise=True)
    
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("127.0.0.1", 8000))
    s.listen()
    conn, addr = s.accept()
    while True:
        data = conn.recv(1024)
        if not data:
            break
        
        conn.send("HTTP/1.1 200 OK\n".encode())
        conn.send("Content-Type: text/html\n\n".encode())
        conn.send("\n".encode())
        conn.send("""
            <html>
                <body>
                    <h1>Auth completed</h1>
                </body>
            </html>
        """.encode())
        conn.close()

        try:
            code = pat.findall(data.decode('utf-8'))[0][0]
            return code
        except:
            break



class MediaEntryMutable:
    
    def __init__(self, id, client):
        self._id = id
        self._client = client

    def set_score(self, score):
        query = """
        mutation ($id: Int, $score: Float) {
            SaveMediaListEntry (id: $id, score: $score) {
                id
                score
            }
        }
        """

        resp = self._client._request(query, vars={
            "id": self._id,
            "score": score
        })

        return "OK"
    
    def delete(self):
        query = """
        mutation ($id: Int) {
            DeleteMediaListEntry (id: $id) {
                deleted
            }
        }
        """

        resp = self._client._request(query, vars={
            "id": self._id,
        })

        if resp.DeleteMediaListEntry.deleted:
            return "OK"
        else:
            return "ERR"

    @classmethod
    def _from_media_id(cls, client, media_id):
        query = """
        mutation ($mediaId: Int) {
            SaveMediaListEntry (mediaId: $mediaId) {
                id
            }
        }
        """

        resp = client._request(query, vars={
            "mediaId": media_id
        })

        return cls(resp.SaveMediaListEntry.id, client)



class Scheme:

    def __init__(self):
        self._layers = []
        self._up = {}

    def __getattr__(self, k):
        return self[k]
    
    def __getitem__(self, k):
        self._layers.append(k)
        return self
    
    def __call__(self, **kwargs):
        d = dict(**kwargs)
        if d != {}:
            self._up[self._layers[-1]] = f"{self._layers[-1]} ({', '.join([f'{k}: {v}' for k, v in d.items()])})"
        return self
    
    @classmethod
    def _construct(cls, *schs):
        d = {k: v for sch in schs for k, v in sch._up.items()}
        return construct_query_as_string(construct_query(list(schs)), d)
    
    @classmethod
    def _combine(cls, scha, schb):
        n = cls()
        n._layers = scha._layers + schb._layers
        n._up = scha._up | schb._up
        return n
    
def construct_query(schs: list[Scheme]):
    temp = {}

    for sch in schs:
        sch = sch._layers
        if not sch[0] in temp:
            temp[sch[0]] = {}
        for i in range(1, len(sch)):
            if not sch[i] in get_nest_list(get_numbers_up_to(i), temp, i, sch):
                get_nest_list(get_numbers_up_to(i), temp, i, sch)[sch[i]] = {}

    
    return temp


def construct_query_as_string(d, l, ss=""):
    if type(d) != dict:
        return str(d)
    
    s = ""

    for k, v in d.items():
        nk = k if k not in l else l[k]
        if v == {}:
            s = f"{s}{ss}{nk}\n"
        else:
            t = construct_query_as_string(v, l, f'{ss}  ')
            t = t[0:-1]
            s = f"{s}{ss}{nk} {{\n{t}\n{ss}}}\n"

    return s


def get_numbers_up_to(i):
    l = []
    n = i
    while n > 0:
        l.append(n)
        n -= 1
    
    return l

def get_nest_list(n, l, m, sch):
    prev = l
    for i in n:
        prev = prev[sch[m-i]]
    return prev
    



import json

from Anilist.auth.token import get_token

class Auth:

    def __init__(self, client_id, client_secret):
        self.__id = client_id
        self.__secret = client_secret
        self.__token = None

        self.__gen_token()

    def __gen_token(self):
        self.__token = get_token(self.__id, self.__secret)

    @property
    def token(self):
        return self.__token

    @classmethod
    def from_config_object(cls, config):
        return cls(config["ID"], config["SECRET"])
    
    @classmethod
    def from_config_file(cls, file):
        with open(file, 'r') as f:
            config = json.loads(f.read())
        return cls(config["ID"], config["SECRET"])



class RequestError(Exception):
    
    def __init__(self, message: str, code: int, locations: list[tuple[int, int]]):
        super().__init__()

        self._message = message
        self._code = code
        self._line = locations[0][0]
        self._column = locations[0][1]

    def __str__(self):
        return f"API Request returned error with code {self._code}. Error at line {self._line}, column {self._column}: '{self._message}' Use DEBUG logging level to view the query."

    @classmethod
    def from_json(cls, json_obj):
        vars = [json_obj["message"], json_obj["status"], [(i["line"], i["column"]) for i in json_obj["locations"]]]
        if "validation" not in json_obj:
            return cls(*vars)
        return MutationRequestError(json_obj["validation"], *vars)

class MutationRequestError(RequestError):
    
    def __init__(self, validation, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._validation = validation



from Anilist import QueryClient, Scheme
import logging
from unittest import TestCase

class BaseTest(TestCase):

    def test_scheme(self):
        pass
        #_ = Scheme._construct(Scheme().A.B(k='$3'), Scheme().A.B.C, Scheme().A.B.C(k="$4").D)
    
    def test_query(self):
        q = QueryClient()

        q.media_list(username="xxxAnn", per_page=100).entries[0]


# define wrappers for Anilist types

class AnilistType:
    def __new__(cls, v):
        if cls == AnilistType:
            if type(v) is str: 
                return AnilistString(None)
            if type(v) is int:
                return AnilistInt(None)
            if issubclass(v, AnilistType):
                return v
        else:
            return super().__new__(cls)

    def __str__(self):
        print(self.__class__)
        return self.__class__.ANILIST_TYPE_NAME
    
    def __repr__(self):
        return str(self)

class AnilistString(AnilistType):
    
    ANILIST_TYPE_NAME = "String"

class AnilistInt(AnilistType):

    ANILIST_TYPE_NAME = "Int"




from Anilist.anilist_types import AnilistType

class Vars:
    
    def __init__(self, **kwargs):
        dic = dict(kwargs)
        l = []
        for k, v in dic.items(): 
            l.append(f"${k}: {AnilistType(v).__class__.ANILIST_TYPE_NAME}")

        self._dic = dic
        self._vars = l

    @property
    def _names(self):
        return ', '.join(self._vars)
    
    @property
    def _json(self):
        return self._dic


from Anilist.scheme import MediaScheme, Scheme
from Anilist.vars.vars import Vars
class MediaListQuery:
    
    def __init__(self, client, username: str, per_page: int=10, starting_page: int = 1, languages=["english"], sizes=["extraLarge"]):
        self._username = username
        self._per_page = per_page
        self._starting_page = starting_page
        self._languages = languages
        self._client = client
        self._sizes = sizes
        self._media_entries = []
        self.DEFAULT_QUERY = [
            MediaScheme().id, 
            *[MediaScheme().title[lang] for lang in self._languages], 
            *[MediaScheme().coverImage[size] for size in self._sizes]
        ]
        self.DEFAULT_VARS = Vars(usr=self._username, page=self._starting_page, perPage=self._per_page)
        # load default values
        self._base_query()

    @property
    def entries(self):
        return self._media_entries
    
    def query(self, *schs, default=True):
    
        if default:
            schs = list(schs)
            schs.extend(self.DEFAULT_QUERY)

        return self._query(*schs)
    
    def _query(self, *schs):

        head_sch = Scheme().Page(page="$page", perPage="$perPage").mediaList(userName="$usr")

        query = self._client._create_query(self.DEFAULT_VARS, *schs, head_sch=head_sch)

        pg = self._starting_page
        
        vals = self.DEFAULT_VARS._json

        temp = []

        while True:
            resp = self._client._request(query, vars=vals)
            data = ([v.media for v in resp.Page.mediaList])

            temp.extend(data)

            if data == []:
                break

            pg += 1
            vals["page"] = pg

        self._media_entries = temp

    def _base_query(self):
        self.query()


    
from Anilist.client import Client
from Anilist.query.media_list import MediaListQuery

import logging

from Anilist.vars.vars import Vars

class QueryClient(Client):

    def __init__(self, level=logging.INFO):
        Client.__init__(self, None, level)

    def _query_type(self):
        return "query"

    def media_list(self, *, username, per_page: int=10, starting_page: int=1, languages=["english"], sizes=["extraLarge"]):
        return MediaListQuery(
            client=self,
            username=username, 
            per_page=per_page, 
            starting_page=starting_page, 
            languages=languages, 
            sizes=sizes
        )



import requests, logging
from Anilist.scheme.scheme import Scheme

from Anilist.utils import AnilistObject, AnilistLogger, consts
from Anilist.errors import RequestError
class Client:

    URI = consts.API_URI
    DEFAULT_HEADERS = consts.DEFAULT_HEADERS

    def __init__(self, auth, level):
        self._auth = auth
        self._headers = None
        _ = AnilistLogger(level)

    def _request(self, query, vars):

        log = AnilistLogger()
        log.info(f"Sending request to URI {self.URI} with headers {self.headers}")
        log.debug(f"The query is {query}\n The variables are {vars}")
        
        req = requests.post(self.URI, json={
            "query": query,
            "variables": vars,
        }, headers = self.headers)

        resp = req.json()

        log.info(f"Received response to request with status code {req.status_code}")

        if "errors" in resp and resp["errors"] != []:
            raise RequestError.from_json(resp["errors"][0])
        
        obj = AnilistObject(resp["data"])
        return obj
        
    def _create_query(self, vars, *schs, head_sch=None):
        schs = list(schs) if head_sch is None else [Scheme._combine(head_sch, sch) for sch in schs]

        query = """
            {} ({}) {{
                {}
            }}
        """.format(self._query_type(), vars._names, Scheme._construct(*schs))

        return query
    
    def _query_type(self):
        return "null"
    
    @property
    def headers(self):
        if self._headers == None:
            self._headers = self._gen_headers()
        
        return self._headers
        
    def _gen_headers(self):
        temp = self.DEFAULT_HEADERS

        if self._auth != None:
            temp["Authorization"] = f"Bearer {self._auth.token}"

        return temp



